ADC1 = 1073815552
ADC1_COMMON = 1073816320
ADC_CR1 = 4
ADC_CR2 = 8
ADC_DR = 76
ADC_HTR = 36
ADC_JDR1 = 60
ADC_JDR2 = 64
ADC_JDR3 = 68
ADC_JDR4 = 72
ADC_JOFR1 = 20
ADC_JOFR2 = 24
ADC_JOFR3 = 28
ADC_JOFR4 = 32
ADC_JSQR = 56
ADC_LTR = 40
ADC_SMPR1 = 12
ADC_SMPR2 = 16
ADC_SQR1 = 44
ADC_SQR2 = 48
ADC_SQR3 = 52
ADC_SR = 0
CRC = 1073885184
CRC_CR = 8
CRC_DR = 0
CRC_IDR = 4
DBGMCU = 3758366720
DBGMCU_APB1FZ = 8
DBGMCU_APB2FZ = 12
DBGMCU_CR = 4
DBGMCU_IDCODE = 0
DMA1 = 1073897472
DMA2 = 1073898496
DMA_HIFCR = 12
DMA_HISR = 4
DMA_LIFCR = 8
DMA_LISR = 0
EXTI = 1073822720
EXTI_EMR = 4
EXTI_FTSR = 12
EXTI_IMR = 0
EXTI_PR = 20
EXTI_RTSR = 8
EXTI_SWIER = 16
FLASH = 1073888256
FLASH_ACR = 0
FLASH_CR = 16
FLASH_KEYR = 4
FLASH_OPTCR = 20
FLASH_OPTCR1 = 24
FLASH_OPTKEYR = 8
FLASH_SR = 12
GPIOA = 1073872896
GPIOB = 1073873920
GPIOC = 1073874944
GPIOD = 1073875968
GPIOE = 1073876992
GPIOH = 1073880064
GPIO_AFR0 = 32
GPIO_AFR1 = 36
GPIO_BSRR = 24
GPIO_BSRRH = 26
GPIO_BSRRL = 24
GPIO_IDR = 16
GPIO_LCKR = 28
GPIO_MODER = 0
GPIO_ODR = 20
GPIO_OSPEEDR = 8
GPIO_OTYPER = 4
GPIO_PUPDR = 12
I2C1 = 1073763328
I2C2 = 1073764352
I2C3 = 1073765376
I2C_CCR = 28
I2C_CR1 = 0
I2C_CR2 = 4
I2C_DR = 16
I2C_FLTR = 36
I2C_OAR1 = 8
I2C_OAR2 = 12
I2C_SR1 = 20
I2C_SR2 = 24
I2C_TRISE = 32
I2S2EXT = 1073755136
I2S3EXT = 1073758208
IWDG = 1073754112
IWDG_KR = 0
IWDG_PR = 4
IWDG_RLR = 8
IWDG_SR = 12
PWR = 1073770496
PWR_CR = 0
PWR_CSR = 4
RCC = 1073887232
RCC_AHB1ENR = 48
RCC_AHB1LPENR = 80
RCC_AHB1RSTR = 16
RCC_AHB2ENR = 52
RCC_AHB2LPENR = 84
RCC_AHB2RSTR = 20
RCC_AHB3ENR = 56
RCC_AHB3LPENR = 88
RCC_AHB3RSTR = 24
RCC_APB1ENR = 64
RCC_APB1LPENR = 96
RCC_APB1RSTR = 32
RCC_APB2ENR = 68
RCC_APB2LPENR = 100
RCC_APB2RSTR = 36
RCC_BDCR = 112
RCC_CFGR = 8
RCC_CIR = 12
RCC_CR = 0
RCC_CSR = 116
RCC_DCKCFGR = 140
RCC_PLLCFGR = 4
RCC_PLLI2SCFGR = 132
RCC_SSCGR = 128
RTC = 1073752064
RTC_ALRMAR = 28
RTC_ALRMBR = 32
RTC_BKP0R = 80
RTC_BKP10R = 120
RTC_BKP11R = 124
RTC_BKP12R = 128
RTC_BKP13R = 132
RTC_BKP14R = 136
RTC_BKP15R = 140
RTC_BKP16R = 144
RTC_BKP17R = 148
RTC_BKP18R = 152
RTC_BKP19R = 156
RTC_BKP1R = 84
RTC_BKP2R = 88
RTC_BKP3R = 92
RTC_BKP4R = 96
RTC_BKP5R = 100
RTC_BKP6R = 104
RTC_BKP7R = 108
RTC_BKP8R = 112
RTC_BKP9R = 116
RTC_CALIBR = 24
RTC_CALR = 60
RTC_CR = 8
RTC_DR = 4
RTC_ISR = 12
RTC_PRER = 16
RTC_SHIFTR = 44
RTC_SSR = 40
RTC_TAFCR = 64
RTC_TR = 0
RTC_TSDR = 52
RTC_TSSSR = 56
RTC_TSTR = 48
RTC_WPR = 36
RTC_WUTR = 20
SDIO = 1073818624
SPI1 = 1073819648
SPI2 = 1073756160
SPI3 = 1073757184
SPI4 = 1073820672
SPI5 = 1073827840
SPI_CR1 = 0
SPI_CR2 = 4
SPI_CRCPR = 16
SPI_DR = 12
SPI_I2SCFGR = 28
SPI_I2SPR = 32
SPI_RXCRCR = 20
SPI_SR = 8
SPI_TXCRCR = 24
SYSCFG = 1073821696
SYSCFG_CMPCR = 32
SYSCFG_EXTICR0 = 8
SYSCFG_EXTICR1 = 12
SYSCFG_EXTICR2 = 16
SYSCFG_EXTICR3 = 20
SYSCFG_MEMRMP = 0
SYSCFG_PMC = 4
TIM1 = 1073807360
TIM10 = 1073824768
TIM11 = 1073825792
TIM2 = 1073741824
TIM3 = 1073742848
TIM4 = 1073743872
TIM5 = 1073744896
TIM9 = 1073823744
TIM_ARR = 44
TIM_BDTR = 68
TIM_CCER = 32
TIM_CCMR1 = 24
TIM_CCMR2 = 28
TIM_CCR1 = 52
TIM_CCR2 = 56
TIM_CCR3 = 60
TIM_CCR4 = 64
TIM_CNT = 36
TIM_CR1 = 0
TIM_CR2 = 4
TIM_DCR = 72
TIM_DIER = 12
TIM_DMAR = 76
TIM_EGR = 20
TIM_OR = 80
TIM_PSC = 40
TIM_RCR = 48
TIM_SMCR = 8
TIM_SR = 16
USART1 = 1073811456
USART2 = 1073759232
USART6 = 1073812480
USART_BRR = 8
USART_CR1 = 12
USART_CR2 = 16
USART_CR3 = 20
USART_DR = 4
USART_GTPR = 24
USART_SR = 0
WWDG = 1073753088
WWDG_CFR = 4
WWDG_CR = 0
WWDG_SR = 8
mem16 = None
mem32 = None
mem8 = None
